/*
 * This file is a part of PikoTools
 * and is distributed under the (new) BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * Neither the name Tomasz Sowa nor the names of contributors to this
 *    project may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef headerfile_picotools_convert_patternreplacer
#define headerfile_picotools_convert_patternreplacer

#include <string>
#include "textstream/textstream.h"
#include "strtoint.h"



namespace PT
{

template<typename CharType, typename StrType>
class PatternReplacerBase
{
public:

	/*
	 * output_string can be the same string as input pattern
	 */
	template<typename ... Types>
	void ReplaceToString(const StrType & pattern, StrType & output_string, Types ... types)
	{
		ReplaceGeneric(pattern, output_string, types...);
	}


	template<typename ... Types>
	StrType Replace(const StrType & pattern, Types ... types)
	{
		StrType output_string;
		ReplaceGeneric(pattern, output_string, types...);

		return output_string;
	}


private:

	std::vector<StrType> params;
	TextStreamBase<CharType, 256, 4096> buffer;
	StrType temp_str;



	template<typename ... Types>
	void ReplaceGeneric(const StrType & pattern, StrType & output_string, Types ... types)
	{
		params.clear();
		AddParams(types...);
		ReplacePattern(pattern);
		buffer.to_string(output_string);

		params.clear();
		buffer.clear();
		temp_str.clear();
	}


	void AddParams()
	{
	}


	template<typename Type, typename ... Types>
	void AddParams(const Type & type, Types ... types)
	{
		buffer.clear();
		buffer << type;
		buffer.to_string(temp_str);
		params.push_back(temp_str);
		temp_str.clear();

		AddParams(types...);
	}


	void ReplacePattern(const StrType & pattern)
	{
		buffer.clear();

		for(size_t i=0 ; i<pattern.size() ; )
		{
			bool pattern_changed = false;

			if( pattern[i] == '%' )
			{
				if( i + 1 < pattern.size() )
				{
					if( pattern[i+1] == '%' )
					{
						buffer << '%';
						i += 2;
						pattern_changed = true;
					}
					else
					{
						const CharType * index_start_str = &pattern[i+1];
						const CharType * after_str;
						bool was_overflow;
						unsigned long index = Toul(index_start_str, 10, &after_str, &was_overflow, false);

						if( !was_overflow && after_str > index_start_str && (size_t)index < params.size() )
						{
							i = i + 1 + (after_str - index_start_str);
							buffer << params[index];
							pattern_changed = true;
						}
					}
				}
			}

			if( !pattern_changed )
			{
				buffer << pattern[i];
				i += 1;
			}
		}
	}


};


typedef PatternReplacerBase<char, std::string>     PatternReplacer;
typedef PatternReplacerBase<wchar_t, std::wstring> WPatternReplacer;


}


#endif
